#!/usr/bin/env python3

# Swap Information's
__version__: str = "0.4.1"
__license__: str = "AGPL-3.0"
__author__: str = "Meheret Tesfaye Batu"
__email__: str = "meherett@zoho.com"
__description__: str = "Cross-chain atomic swap between the networks of two cryptocurrencies " \
                       "based on Hash Time Lock Contracts (HTLC's) protocol."

__all__: list = [
    "__version__", "__license__", "__author__", "__email__", "__description__"
]
